from __future__ import absolute_import, division, print_function

import datetime

import numpy as np
from pandas import DataFrame, Series
from datashape import to_numpy, to_numpy_dtype
from numbers import Number

from ..expr import (
    Reduction, Field, Projection, Broadcast, Selection, ndim,
    Distinct, Sort, Tail, Head, Label, ReLabel, Expr, Slice, Join,
    std, var, count, nunique, Summary, IsIn,
    BinOp, UnaryOp, USub, Not, nelements, Repeat, Concat, Interp,
    UTCFromTimestamp, DateTimeTruncate,
    Transpose, TensorDot, Coerce, isnan,
    greatest, least, BinaryMath, atan2, Coalesce, Cast
)
from ..utils import keywords

from .core import base, compute
from .pandas import array_coalesce
from ..dispatch import dispatch
from odo import into
import pandas as pd

__all__ = ['np']


@dispatch(Field, np.ndarray)
def compute_up(c, x, **kwargs):
    if x.dtype.names and c._name in x.dtype.names:
        return x[c._name]
    if not x.dtype.names and x.shape[1] == len(c._child.fields):
        return x[:, c._child.fields.index(c._name)]
    raise NotImplementedError()  # pragma: no cover


@dispatch(Projection, np.ndarray)
def compute_up(t, x, **kwargs):
    if x.dtype.names and all(col in x.dtype.names for col in t.fields):
        return x[t.fields]
    if not x.dtype.names and x.shape[1] == len(t._child.fields):
        return x[:, [t._child.fields.index(col) for col in t.fields]]
    raise NotImplementedError()  # pragma: no cover


try:
    from .numba import broadcast_numba as broadcast_ndarray
except ImportError:
    def broadcast_ndarray(t, *data, **kwargs):
        del kwargs['scope']
        d = dict(zip(t._scalar_expr._leaves(), data))
        return compute(t._scalar_expr, d, return_type='native', **kwargs)


compute_up.register(Broadcast, np.ndarray)(broadcast_ndarray)
for i in range(2, 6):
    compute_up.register(Broadcast, *([(np.ndarray, Number)] * i))(broadcast_ndarray)


@dispatch(Repeat, np.ndarray)
def compute_up(t, data, _char_mul=np.char.multiply, **kwargs):
    if isinstance(t.lhs, Expr):
        return _char_mul(data, t.rhs)
    else:
        return _char_mul(t.lhs, data)


@compute_up.register(Repeat, np.ndarray, (np.ndarray, base))
@compute_up.register(Repeat, base, np.ndarray)
def compute_up_np_repeat(t, lhs, rhs, _char_mul=np.char.multiply, **kwargs):
    return _char_mul(lhs, rhs)


def _interp(arr, v, _Series=pd.Series, _charmod=np.char.mod):
    """
    Delegate to the most efficient string formatting technique based on
    the length of the array.
    """
    if len(arr) >= 145:
        return _Series(arr) % v

    return _charmod(arr, v)


@dispatch(Interp, np.ndarray)
def compute_up(t, data, **kwargs):
    if isinstance(t.lhs, Expr):
        return _interp(data, t.rhs)
    else:
        return _interp(t.lhs, data)


@compute_up.register(Interp, np.ndarray, (np.ndarray, base))
@compute_up.register(Interp, base, np.ndarray)
def compute_up_np_interp(t, lhs, rhs, **kwargs):
    return _interp(lhs, rhs)


@compute_up.register(greatest, np.ndarray, (np.ndarray, base))
@compute_up.register(greatest, base, np.ndarray)
def compute_up_greatest(expr, lhs, rhs, **kwargs):
    return np.maximum(lhs, rhs)


@compute_up.register(least, np.ndarray, (np.ndarray, base))
@compute_up.register(least, base, np.ndarray)
def compute_up_least(expr, lhs, rhs, **kwargs):
    return np.minimum(lhs, rhs)


@dispatch(BinOp, np.ndarray, (np.ndarray, base))
def compute_up(t, lhs, rhs, **kwargs):
    return t.op(lhs, rhs)


@dispatch(BinOp, base, np.ndarray)
def compute_up(t, lhs, rhs, **kwargs):
    return t.op(lhs, rhs)


@dispatch(BinOp, np.ndarray)
def compute_up(t, data, **kwargs):
    if isinstance(t.lhs, Expr):
        return t.op(data, t.rhs)
    else:
        return t.op(t.lhs, data)


@compute_up.register(BinaryMath, np.ndarray, (np.ndarray, base))
@compute_up.register(BinaryMath, base, np.ndarray)
def compute_up_binary_math(t, lhs, rhs, **kwargs):
    return getattr(np, type(t).__name__)(lhs, rhs)


@dispatch(BinaryMath, np.ndarray)
def compute_up(t, data, **kwargs):
    func = getattr(np, type(t).__name__)
    if isinstance(t.lhs, Expr):
        return func(data, t.rhs)
    else:
        return func(t.lhs, data)


@compute_up.register(atan2, np.ndarray, (np.ndarray, base))
@compute_up.register(atan2, base, np.ndarray)
def compute_up_binary_math(t, lhs, rhs, **kwargs):
    return np.arctan2(lhs, rhs)


@dispatch(atan2, np.ndarray)
def compute_up(t, data, **kwargs):
    if isinstance(t.lhs, Expr):
        return np.arctan2(data, t.rhs)
    else:
        return np.arctan2(t.lhs, data)


@dispatch(UnaryOp, np.ndarray)
def compute_up(t, x, **kwargs):
    return getattr(np, t.symbol)(x)


@dispatch(Not, np.ndarray)
def compute_up(t, x, **kwargs):
    return np.logical_not(x)


@dispatch(USub, np.ndarray)
def compute_up(t, x, **kwargs):
    return np.negative(x)


inat = np.datetime64('NaT').view('int64')


@dispatch(count, np.ndarray)
def compute_up(t, x, **kwargs):
    result_dtype = to_numpy_dtype(t.dshape)
    if issubclass(x.dtype.type, (np.floating, np.object_)):
        return pd.notnull(x).sum(keepdims=t.keepdims, axis=t.axis,
                                 dtype=result_dtype)
    elif issubclass(x.dtype.type, np.datetime64):
        return (x.view('int64') != inat).sum(keepdims=t.keepdims, axis=t.axis,
                                             dtype=result_dtype)
    else:
        return np.ones(x.shape, dtype=result_dtype).sum(keepdims=t.keepdims,
                                                        axis=t.axis,
                                                        dtype=result_dtype)


@dispatch(nunique, np.ndarray)
def compute_up(t, x, **kwargs):
    assert t.axis == tuple(range(ndim(t._child)))
    result = len(np.unique(x))
    if t.keepdims:
        result = np.array([result])
    return result


@dispatch(Reduction, np.ndarray)
def compute_up(t, x, **kwargs):
    # can't use the method here, as they aren't Python functions
    reducer = getattr(np, t.symbol)
    if 'dtype' in keywords(reducer):
        return reducer(x, axis=t.axis, keepdims=t.keepdims,
                       dtype=to_numpy_dtype(t.schema))
    return reducer(x, axis=t.axis, keepdims=t.keepdims)


def axify(expr, axis, keepdims=False):
    """ inject axis argument into expression

    Helper function for compute_up(Summary, np.ndarray)

    >>> from blaze import symbol
    >>> s = symbol('s', '10 * 10 * int')
    >>> expr = s.sum()
    >>> axify(expr, axis=0)
    sum(s, axis=(0,))
    """
    return type(expr)(expr._child, axis=axis, keepdims=keepdims)


@dispatch(Summary, np.ndarray)
def compute_up(expr, data, **kwargs):
    shape, dtype = to_numpy(expr.dshape)
    if shape:
        result = np.empty(shape=shape, dtype=dtype)
        for n, v in zip(expr.names, expr.values):
            result[n] = compute(
                axify(v, expr.axis, expr.keepdims),
                data,
                return_type='native',
            )
        return result
    else:
        return tuple(
            compute(axify(v, expr.axis), data, return_type='native')
            for v in expr.values
        )


@dispatch((std, var), np.ndarray)
def compute_up(t, x, **kwargs):
    return getattr(x, t.symbol)(ddof=t.unbiased, axis=t.axis,
            keepdims=t.keepdims)


@compute_up.register(Distinct, np.recarray)
def recarray_distinct(t, rec, **kwargs):
    return pd.DataFrame.from_records(rec).drop_duplicates(
        subset=t.on or None).to_records(index=False).astype(rec.dtype)


@dispatch(Distinct, np.ndarray)
def compute_up(t, arr, _recarray_distinct=recarray_distinct, **kwargs):
    if t.on:
        if getattr(arr.dtype, 'names', None) is not None:
            return _recarray_distinct(t, arr, **kwargs).view(np.ndarray)
        else:
            raise ValueError('malformed expression: no columns to distinct on')

    return np.unique(arr)


@dispatch(Sort, np.ndarray)
def compute_up(t, x, **kwargs):
    if x.dtype.names is None:  # not a struct array
        result = np.sort(x)
    elif (t.key in x.dtype.names or  # struct array
        isinstance(t.key, list) and all(k in x.dtype.names for k in t.key)):
        result = np.sort(x, order=t.key)
    elif t.key:
        raise NotImplementedError("Sort key %s not supported" % t.key)

    if not t.ascending:
        result = result[::-1]

    return result


@dispatch(Head, np.ndarray)
def compute_up(t, x, **kwargs):
    return x[:t.n]


@dispatch(Tail, np.ndarray)
def compute_up(t, x, **kwargs):
    return x[-t.n:]


@dispatch(Label, np.ndarray)
def compute_up(t, x, **kwargs):
    return np.array(x, dtype=[(t.label, x.dtype.type)])


@dispatch(ReLabel, np.ndarray)
def compute_up(t, x, **kwargs):
    types = [x.dtype[i] for i in range(len(x.dtype))]
    return np.array(x, dtype=list(zip(t.fields, types)))


@dispatch(Selection, np.ndarray)
def compute_up(sel, x, **kwargs):
    predicate = compute(sel.predicate, {sel._child: x}, return_type='native')
    cond = getattr(predicate, 'values', predicate)
    return x[cond]


@dispatch(Selection, np.ndarray, np.ndarray)
def compute_up(expr, arr, predicate, **kwargs):
    return arr[predicate]


@dispatch(Selection, np.ndarray, Series)
def compute_up(expr, arr, predicate, **kwargs):
    return arr[predicate.values]


@dispatch(UTCFromTimestamp, np.ndarray)
def compute_up(expr, data, **kwargs):
    return (data * 1e6).astype('datetime64[us]')


@dispatch(Slice, np.ndarray)
def compute_up(expr, x, **kwargs):
    return x[expr.index]


@dispatch(Cast, np.ndarray)
def compute_up(t, x, **kwargs):
    # resolve ambiguity
    return x


@dispatch(Expr, np.ndarray)
def compute_up(t, x, **kwargs):
    ds = t._child.dshape
    if x.ndim > 1 or isinstance(x, np.recarray) or x.dtype.fields is not None:
        return compute_up(t, into(DataFrame, x, dshape=ds), **kwargs)
    else:
        return compute_up(t, into(Series, x, dshape=ds), **kwargs)


@dispatch(nelements, np.ndarray)
def compute_up(expr, data, **kwargs):
    axis = expr.axis
    if expr.keepdims:
        shape = tuple(data.shape[i] if i not in axis else 1
                                    for i in range(ndim(expr._child)))
    else:
        shape = tuple(data.shape[i] for i in range(ndim(expr._child))
                      if i not in axis)
    value = np.prod([data.shape[i] for i in axis])
    result = np.empty(shape)
    result.fill(value)
    result = result.astype('int64')

    return result



# Note the use of 'week': 'M8[D]' here.

# We truncate week offsets "manually" in the compute_up implementation by first
# converting to days then multiplying our measure by 7 this simplifies our code
# by only requiring us to calculate the week offset relative to the day of week.

precision_map = {'year': 'M8[Y]',
                 'month': 'M8[M]',
                 'week': 'M8[D]',
                 'day': 'M8[D]',
                 'hour': 'M8[h]',
                 'minute': 'M8[m]',
                 'second': 'M8[s]',
                 'millisecond': 'M8[ms]',
                 'microsecond': 'M8[us]',
                 'nanosecond': 'M8[ns]'}


# these offsets are integers in units of their representation

epoch = datetime.datetime(1970, 1, 1)
offsets = {
    'week': epoch.isoweekday(),
    'day': epoch.toordinal() # number of days since *Python's* epoch (01/01/01)
}


@dispatch(DateTimeTruncate, (np.ndarray, np.datetime64))
def compute_up(expr, data, **kwargs):
    np_dtype = precision_map[expr.unit]
    offset = offsets.get(expr.unit, 0)
    measure = expr.measure * 7 if expr.unit == 'week' else expr.measure
    result = (((data.astype(np_dtype)
                    .view('int64')
                    + offset)
                    // measure
                    * measure
                    - offset)
                    .astype(np_dtype))
    return result


@dispatch(isnan, np.ndarray)
def compute_up(expr, data, **kwargs):
    return np.isnan(data)


@dispatch(np.ndarray)
def chunks(x, chunksize=1024):
    start = 0
    n = len(x)
    while start < n:
        yield x[start:start + chunksize]
        start += chunksize


@dispatch(Transpose, np.ndarray)
def compute_up(expr, x, **kwargs):
    return np.transpose(x, axes=expr.axes)


@dispatch(TensorDot, np.ndarray, np.ndarray)
def compute_up(expr, lhs, rhs, **kwargs):
    return np.tensordot(lhs, rhs, axes=[expr._left_axes, expr._right_axes])


@dispatch(IsIn, np.ndarray)
def compute_up(expr, data, **kwargs):
    return np.in1d(data, tuple(expr._keys))


@compute_up.register(Join, DataFrame, np.ndarray)
@compute_up.register(Join, np.ndarray, DataFrame)
@compute_up.register(Join, np.ndarray, np.ndarray)
def join_ndarray(expr, lhs, rhs, **kwargs):
    if isinstance(lhs, np.ndarray):
        lhs = DataFrame(lhs)
    if isinstance(rhs, np.ndarray):
        rhs = DataFrame(rhs)
    return compute_up(expr, lhs, rhs, **kwargs)


@dispatch(Coerce, np.ndarray)
def compute_up(expr, data, **kwargs):
    return data.astype(to_numpy_dtype(expr.schema))


@dispatch(Concat, np.ndarray, np.ndarray)
def compute_up(expr, lhs, rhs, _concat=np.concatenate, **kwargs):
    return _concat((lhs, rhs), axis=expr.axis)


compute_up.register(Coalesce, np.ndarray, (np.ndarray, base))(array_coalesce)
compute_up.register(Coalesce, base, np.ndarray)(array_coalesce)


@dispatch(Coalesce, np.ndarray)
def compute_up(t, data, **kwargs):
    if isinstance(t.lhs, Expr):
        lhs = data
        rhs = t.rhs
    else:
        lhs = t.lhs
        rhs = data

    return array_coalesce(t, lhs, rhs)



from __future__ import absolute_import, division, print_function

import pandas as pd
from ..expr import (Expr, Symbol, Field, Arithmetic, UnaryMath, BinaryMath,
                    Date, Time, DateTime, Millisecond, Microsecond, broadcast,
                    sin, cos, Map, UTCFromTimestamp, DateTimeTruncate, symbol,
                    USub, Not, notnull, greatest, least, atan2, Like)
from ..expr import math as expr_math
from ..expr.expressions import valid_identifier
from ..dispatch import dispatch
from . import pydatetime
import numpy as np
import datetime
import fnmatch
import math
import toolz
import itertools


funcnames = ('func_%d' % i for i in itertools.count())


def parenthesize(s):
    if ' ' in s:
        return '(%s)' % s
    else:
        return s


def print_python(leaves, expr):
    """ Print expression to be evaluated in Python

    >>> from blaze.expr import ceil, sin

    >>> t = symbol('t', '{x: int, y: int, z: int, when: datetime}')
    >>> print_python([t], t.x + t.y)
    ('t[0] + t[1]', {})

    Supports mathematical and datetime access

    >>> print_python([t], sin(t.x) > ceil(t.y))  # doctest: +SKIP
    ('math.sin(t[0]) > math.ceil(t[1])', {'math':<module 'math'>})
    >>> print_python([t], t.when.day + 1)
    ('t[3].day + 1', {})

    Specify leaves of the expression to control level of printing

    >>> print_python([t.x, t.y], t.x + t.y)
    ('x + y', {})

    Returns
    -------

    s: string
       A evalable string
    scope: dict
       A namespace to add to be given to eval
    """
    if isinstance(expr, Expr) and any(expr.isidentical(lf) for lf in leaves):
        return valid_identifier(expr._name), {}
    return _print_python(expr, leaves=leaves)


@dispatch(object)
def _print_python(expr, leaves=None):
    return repr(expr), {}


@dispatch((datetime.datetime, datetime.date))
def _print_python(expr, leaves=None):
    return repr(expr), {'datetime': datetime, 'Timestamp': pd.Timestamp}


@dispatch(Symbol)
def _print_python(expr, leaves=None):
    return valid_identifier(expr._name), {}


@dispatch(Field)
def _print_python(expr, leaves=None):
    child, scope = print_python(leaves, expr._child)
    index = expr._child.fields.index(expr._name)
    return '%s[%d]' % (parenthesize(child), index), scope


@dispatch(Arithmetic)
def _print_python(expr, leaves=None):
    lhs, left_scope = print_python(leaves, expr.lhs)
    rhs, right_scope = print_python(leaves, expr.rhs)
    return ('%s %s %s' % (parenthesize(lhs),
                          expr.symbol,
                          parenthesize(rhs)),
            toolz.merge(left_scope, right_scope))


@dispatch(USub)
def _print_python(expr, leaves=None):
    child, scope = print_python(leaves, expr._child)
    return '%s%s' % (expr.symbol, parenthesize(child)), scope


@dispatch(Not)
def _print_python(expr, leaves=None):
    child, scope = print_python(leaves, expr._child)
    return 'not %s' % parenthesize(child), scope


@dispatch(UnaryMath)
def _print_python(expr, leaves=None):
    child, scope = print_python(leaves, expr._child)
    return ('np.%s(%s)' % (type(expr).__name__, child),
            toolz.merge(scope, {'np': np}))


@dispatch(BinaryMath)
def _print_python(expr, leaves=None):
    lhs, scope_lhs = print_python(leaves, expr.lhs)
    rhs, scope_rhs = print_python(leaves, expr.rhs)
    return ('np.%s(%s, %s)' % (type(expr).__name__, lhs, rhs),
            toolz.merge(scope_lhs, scope_rhs, {'np': np}))


@dispatch(atan2)
def _print_python(expr, leaves=None):
    lhs, scope_lhs = print_python(leaves, expr.lhs)
    rhs, scope_rhs = print_python(leaves, expr.rhs)
    return ('np.arctan2(%s, %s)' % (lhs, rhs),
            toolz.merge(scope_lhs, scope_rhs, {'np': np}))


@dispatch(greatest)
def _print_python(expr, leaves=None):
    lhs, scope_lhs = print_python(leaves, expr.lhs)
    rhs, scope_rhs = print_python(leaves, expr.rhs)
    return 'max(%s, %s)' % (lhs, rhs), toolz.merge(scope_lhs, scope_rhs)


@dispatch(least)
def _print_python(expr, leaves=None):
    lhs, scope_lhs = print_python(leaves, expr.lhs)
    rhs, scope_rhs = print_python(leaves, expr.rhs)
    return 'min(%s, %s)' % (lhs, rhs), toolz.merge(scope_lhs, scope_rhs)


@dispatch(expr_math.abs)
def _print_python(expr, leaves=None):
    child, scope = print_python(leaves, expr._child)
    return ('abs(%s)' % child, scope)


@dispatch(Date)
def _print_python(expr, leaves=None):
    child, scope = print_python(leaves, expr._child)
    return ('%s.date()' % parenthesize(child), scope)


@dispatch(Time)
def _print_python(expr, leaves=None):
    child, scope = print_python(leaves, expr._child)
    return ('%s.time()' % parenthesize(child), scope)


@dispatch(Millisecond)
def _print_python(expr, leaves=None):
    child, scope = print_python(leaves, expr._child)
    return ('%s.microsecond // 1000' % parenthesize(child), scope)


@dispatch(UTCFromTimestamp)
def _print_python(expr, leaves=None):
    child, scope = print_python(leaves, expr._child)
    return ('datetime.datetime.utcfromtimestamp(%s)' % parenthesize(child),
            toolz.merge({'datetime': datetime}, scope))


@dispatch(DateTime)
def _print_python(expr, leaves=None):
    child, scope = print_python(leaves, expr._child)
    attr = type(expr).__name__.lower()
    return ('%s.%s' % (parenthesize(child), attr), scope)


@dispatch(DateTimeTruncate)
def _print_python(expr, leaves=None):
    child, scope = print_python(leaves, expr._child)
    scope['truncate'] = pydatetime.truncate
    return ('truncate(%s, %s, "%s")' % (child, expr.measure, expr.unit),
            scope)


@dispatch(Map)
def _print_python(expr, leaves=None):
    child, scope = print_python(leaves, expr._child)
    funcname = next(funcnames)
    return ('%s(%s)' % (funcname, child),
            toolz.assoc(scope, funcname, expr.func))


@dispatch(notnull)
def _print_python(expr, leaves=None):
    child, scope = print_python(leaves, expr._child)
    return ('notnull(%s)' % child,
            toolz.merge(scope, dict(notnull=lambda x: x is not None)))


@dispatch(Like)
def _print_python(expr, leaves):
    child, scope = print_python(leaves, expr._child)
    return (
        'fnmatch(%s, %r)' % (child, expr.pattern),
        toolz.merge(scope, dict(fnmatch=fnmatch.fnmatch))
    )


@dispatch(Expr)
def _print_python(expr, leaves=None):
    raise NotImplementedError("Do not know how to write expressions of type %s"
                              " to Python code" % type(expr).__name__)


def funcstr(leaves, expr):
    """ Lambda string for an expresion

    >>> t = symbol('t', '{x: int, y: int, z: int, when: datetime}')

    >>> funcstr([t], t.x + t.y)
    ('lambda t: t[0] + t[1]', {})

    >>> funcstr([t.x, t.y], t.x + t.y)
    ('lambda x, y: x + y', {})

    Also returns scope for libraries like math or datetime

    >>> funcstr([t.x, t.y], sin(t.x) + t.y)  # doctest: +SKIP
    ('lambda x, y: math.sin(x) + y', {'math': <module 'math'>})

    >>> from datetime import date
    >>> funcstr([t.x, t.y, t.when], t.when.date > date(2001, 12, 25)) #doctest: +SKIP
    ('lambda x, y, when: when.day > datetime.date(2001, 12, 25)', {'datetime': <module 'datetime'>})
    """
    result, scope = print_python(leaves, expr)

    leaf_names = [print_python([leaf], leaf)[0] for leaf in leaves]

    return 'lambda %s: %s' % (', '.join(leaf_names),
                              result), scope


def lambdify(leaves, expr):
    """ Lambda for an expresion

    >>> t = symbol('t', '{x: int, y: int, z: int, when: datetime}')
    >>> f = lambdify([t], t.x + t.y)
    >>> f((1, 10, 100, ''))
    11

    >>> f = lambdify([t.x, t.y, t.z, t.when], t.x + cos(t.y))
    >>> f(1, 0, 100, '')
    2.0
    """
    s, scope = funcstr(leaves, expr)
    return eval(s, scope)



from __future__ import absolute_import, division, print_function

import sys
from functools import partial
import json as json_module

from datashape.predicates import iscollection, istabular, isscalar
from odo import odo
import pandas as pd
from pandas.compat import BytesIO
from pandas.io.packers import (
    unpack as msgpack_unpack,
    decode as pd_msgpack_object_hook,
    encode as pd_msgpack_default,
)
import pandas.msgpack as msgpack_module
from toolz import identity

from blaze.compatibility import pickle as pickle_module, unicode, PY2
from blaze.interactive import coerce_scalar
from .json_dumps import json_dumps
from .json_dumps_trusted import json_dumps_trusted
from .object_hook import object_hook
from .object_hook_trusted import object_hook_trusted


class SerializationFormat(object):
    """A serialization format for the blaze server and blaze client.

    Parameters
    ----------
    name : str
        The name of the format. This is used on the mediatype to select
        the proper format.
    loads : callable[bytes -> any]
        The function that loads python objects out of the serialized data.
    dumps : callable[any -> bytes]
        The function that serializes python objects to some other format.
    data_loads : callable[any -> any], optional
    data_dumps : callable[any -> any], optional
        Specialized functions for loading and writing only the 'data' field of
        blaze server responses. This allows us to define a more efficient
        serialization format for moving large amounts of data while still
        having a rich representation for the rest of the metadata.
    materialize : callable[(any, DataShape, **kwargs) -> any], optional
        The function used to materialize the result of compute into a form
        suitable for serialization.
    """
    def __init__(self,
                 name,
                 loads,
                 dumps,
                 data_loads=None,
                 data_dumps=None,
                 materialize=None):
        self.name = name
        self.loads = loads
        self.dumps = dumps
        self.data_loads = identity if data_loads is None else data_loads
        self.data_dumps = identity if data_dumps is None else data_dumps
        self.materialize = (
            default_materialize
            if materialize is None else
            materialize
        )

    def __repr__(self):
        return '<%s: %r>' % (type(self).__name__, self.name)
    __str__ = __repr__


def default_materialize(data, dshape, odo_kwargs):
    if iscollection(dshape):
        return odo(data, list, **odo_kwargs)
    if isscalar(dshape):
        return coerce_scalar(data, str(dshape), odo_kwargs)

    return data


def _coerce_str(bytes_or_str):
    if isinstance(bytes_or_str, unicode):
        return bytes_or_str
    return bytes_or_str.decode('utf-8')


json = SerializationFormat(
    'json',
    lambda data: json_module.loads(_coerce_str(data), object_hook=object_hook),
    partial(json_module.dumps, default=json_dumps),
)


json_trusted = SerializationFormat(
        'json_trusted',
        lambda data: json_module.loads(_coerce_str(data), object_hook=object_hook_trusted),
        partial(json_module.dumps, default=json_dumps_trusted)
)


pickle = SerializationFormat(
    'pickle',
    pickle_module.loads,
    partial(pickle_module.dumps, protocol=pickle_module.HIGHEST_PROTOCOL),
)

msgpack = SerializationFormat(
    'msgpack',
    partial(msgpack_module.unpackb, encoding='utf-8', object_hook=object_hook),
    partial(msgpack_module.packb, default=json_dumps),
)

msgpack_trusted = SerializationFormat(
    'msgpack_trusted',
    partial(msgpack_module.unpackb, encoding='utf-8', object_hook=object_hook_trusted),
    partial(msgpack_module.packb, default=json_dumps_trusted),
)


try:
    import blosc
    del blosc
    compress = 'blosc'
except ImportError:
    compress = None


def _final_base(values):
    base = values.base
    while True:
        try:
            base = base.base
        except AttributeError:
            return base


def fastmsgpack_object_hook(ob):
    typ = ob.get('typ')
    if typ is None:
        return ob
    if typ == 'nat':
        return pd.NaT
    if typ == 'block_manager':
        df = pd_msgpack_object_hook(ob)
        for block in df._data.blocks:
            values = block.values
            if not values.flags.writeable:
                if sys.getrefcount(_final_base(values)) == 2:
                    # Compatibility shim until this is fixed in pandas
                    # See: https://github.com/pydata/pandas/pull/12359
                    # Mark that we own the memory for this block.
                    # We know that we own it because the only 2 references
                    # for the base object are the values of this block and
                    # TOS because we are passing this object to a function.
                    values.flags.writeable = True
                else:
                    # If we cannot 'safely' make the bytes object mutable
                    # then just copy the data.
                    block.values = values.copy()
        return df
    return pd_msgpack_object_hook(ob)


if PY2:
    def _l1(bs):
        if isinstance(bs, unicode):
            return bs.encode('latin1')
        return bs

    def _str(bs):
        if isinstance(bs, str):
            return bs.decode('utf-8')
        return bs
else:
    def _l1(bs):
        return bs

    def _str(bs):
        return bs


def fastmsgpack_data_loads(data):
    raw = list(msgpack_unpack(
        BytesIO(_l1(data)),
        object_hook=fastmsgpack_object_hook,
    ))
    # raw will always be a list, which is most likely a list containing
    # a single dataframe or series
    if len(raw) == 1:
        # we only serialized one structure, just return it
        return raw[0]
    return raw


def fastmsgpack_loads(data, object_hook=object_hook):
    raw = list(msgpack_unpack(
        BytesIO(_l1(data)),
        object_hook=object_hook,
    ))
    # raw will always be a list, which is most likely a list containing
    # a single dataframe or series
    if len(raw) == 1:
        # we only serialized one structure, just return it
        return raw[0]
    return raw


def fastmsgpack_default(ob):
    # Compatibility shim until this is fixed in pandas
    # See: https://github.com/pydata/pandas/pull/12307
    if ob is pd.NaT:
        return {'typ': 'nat'}
    return pd_msgpack_default(ob)


def fastmsgpack_data_dumps(data):
    return {
        '__!bytes': pd.to_msgpack(
            None,
            data,
            compress=compress,
            default=fastmsgpack_default,
            encoding='utf-8',
            use_bin_type=True,
        ),
    }


def fastmsgpack_dumps(data, default=json_dumps):
    return pd.to_msgpack(
        None,
        data,
        compress=compress,
        default=default,
        encoding='utf-8',
        use_bin_type=True,
    )


def fastmsgpack_materialize(data, dshape, odo_kwargs):
    if istabular(dshape):
        return odo(data, pd.DataFrame, **odo_kwargs)
    if iscollection(dshape):
        return odo(data, pd.Series, **odo_kwargs)
    if isscalar(dshape):
        return coerce_scalar(data, str(dshape), odo_kwargs)
    return data


fastmsgpack = SerializationFormat(
    'fastmsgpack',
    loads=fastmsgpack_loads,
    dumps=fastmsgpack_dumps,
    data_loads=fastmsgpack_data_loads,
    data_dumps=fastmsgpack_data_dumps,
    materialize=fastmsgpack_materialize,
)


fastmsgpack_trusted = SerializationFormat(
    'fastmsgpack_trusted',
    loads=partial(fastmsgpack_loads, object_hook=object_hook_trusted),
    dumps=partial(fastmsgpack_dumps, default=json_dumps_trusted),
    data_loads=fastmsgpack_data_loads,
    data_dumps=fastmsgpack_data_dumps,
    materialize=fastmsgpack_materialize,
)


all_formats = frozenset(g for g in globals().values()
                        if isinstance(g, SerializationFormat))
trusted_formats = frozenset([json_trusted,
                             msgpack_trusted,
                             fastmsgpack_trusted])

__all__ = ['SerializationFormat', 'all_formats', 'trusted_formats'] + list(f.name for f in all_formats)



"""
Expression splitting for chunked computation

To evaluate an expression on a large dataset we may need to chunk that dataset
into pieces and evaluate on each of the pieces individually.  This module
contains logic to break up an expression-to-be-evaluated-on-the-entire-array
into

1.  An expression to be evaluated on each chunk of the data
2.  An expression to be evaluated on the concatenated intermediate results

As an example, consider counting the number of non-null elements in a large
list.  We have the following recipe

1.  Break the large collection into chunks, each of which fits comfortably into
    memory
2.  For each chunk, call compute chunk.count()
3.  Gather all of these results into a single list (which hopefully fits in
    memory)
4.  for this aggregate, compute aggregate.sum()

And so, given this expression

    expr -> expr.count()

We needed the following expressions

    chunk -> chunk.count()
    agg -> agg.sum()

This module performs this transformation for a wide array of chunkable
expressions.  It supports elementwise operations, reductions,
split-apply-combine, and selections.  It notably does not support sorting,
joining, or slicing.

If explicit chunksizes are given it can also reason about the size and shape of
the intermediate aggregate.  It can also do this in N-Dimensions.
"""
from __future__ import absolute_import, division, print_function

from toolz import concat as tconcat
import datashape
from datashape.predicates import isscalar, isrecord
from math import floor

from .core import *
from .expressions import *
from .strings import Like
from .expressions import shape
from .math import sqrt
from .reductions import *
from .split_apply_combine import *
from .collections import *
from ..dispatch import dispatch
from ..compatibility import builtins

good_to_split = (Reduction, Summary, By, Distinct)
can_split = good_to_split + (Like, Selection, ElemWise, Apply)

__all__ = ['path_split', 'split']

def path_split(leaf, expr):
    """ Find the right place in the expression tree/line to parallelize

    >>> t = symbol('t', 'var * {name: string, amount: int, id: int}')

    >>> path_split(t, t.amount.sum() + 1)
    sum(t.amount)

    >>> path_split(t, t.amount.distinct().sort())
    distinct(t.amount)
    """
    last = None
    for node in list(path(expr, leaf))[:-1][::-1]:
        if isinstance(node, good_to_split):
            return node
        elif not isinstance(node, can_split):
            return last
        last = node
    return node


def split(leaf, expr, chunk=None, agg=None, **kwargs):
    """ Split expression for chunked computation

    Break up a computation ``leaf -> expr`` so that it can be run in chunks.
    This returns two computations, one to perform on each chunk and then one to
    perform on the union of these intermediate results

        chunk -> chunk-expr
        aggregate -> aggregate-expr

    The chunk_expr will have the same dimesions as the input
    (reductions set keepdims=True) so that this should work cleanly with
    concatenation functions like ``np.concatenate``.

    Returns
    -------

    Pair of (Symbol, Expr) pairs

        (chunk, chunk_expr), (aggregate, aggregate_expr)

    >>> t = symbol('t', 'var * {name: string, amount: int, id: int}')
    >>> expr = t.id.count()
    >>> split(t, expr)
    ((<`chunk` symbol; dshape='...'>, count(chunk.id, keepdims=True)), (<`aggregate` symbol; dshape='...'>, sum(aggregate)))
    """
    center = path_split(leaf, expr)
    if chunk is None:
        if leaf.ndim > 1:
            raise ValueError("Please provide a chunk symbol")
        else:
            chunk = symbol('chunk', datashape.var * leaf.dshape.measure)

    chunk_expr = _split_chunk(center, leaf=leaf, chunk=chunk, **kwargs)
    chunk_expr_with_keepdims = _split_chunk(center, leaf=leaf, chunk=chunk,
                                            keepdims=True)

    if agg is None:
        agg_shape = aggregate_shape(leaf, expr, chunk, chunk_expr_with_keepdims)
        agg_dshape = DataShape(*(agg_shape + (chunk_expr.dshape.measure,)))
        agg = symbol('aggregate', agg_dshape)

    agg_expr = _split_agg(center, leaf=leaf, agg=agg)

    return ((chunk, chunk_expr),
            (agg, expr._subs({center: agg})._subs({agg: agg_expr})))


reductions = {sum: (sum, sum), count: (count, sum),
              min: (min, min), max: (max, max),
              any: (any, any), all: (all, all),
              nelements: (nelements, sum)}

@dispatch(Expr)
def _split(expr, leaf=None, chunk=None, agg=None, keepdims=True):
    return ((chunk, _split_chunk(expr, leaf=leaf, chunk=chunk,
                                 keepdims=keepdims)),
            (agg, _split_agg(expr, leaf=leaf, agg=agg)))


@dispatch(tuple(reductions))
def _split_chunk(expr, leaf=None, chunk=None, keepdims=True):
    a, b = reductions[type(expr)]
    return a(expr._subs({leaf: chunk})._child,
             keepdims=keepdims,
             axis=expr.axis)

@dispatch(tuple(reductions))
def _split_agg(expr, leaf=None, agg=None):
    a, b = reductions[type(expr)]
    return b(agg, axis=expr.axis, keepdims=expr.keepdims)


@dispatch(mean)
def _split_chunk(expr, leaf=None, chunk=None, keepdims=True):
    child = expr._subs({leaf: chunk})._child
    return summary(total=child.sum(), count=child.count(), keepdims=keepdims,
            axis=expr.axis)

@dispatch(mean)
def _split_agg(expr, leaf=None, agg=None):
    total = agg.total.sum(axis=expr.axis, keepdims=expr.keepdims)
    count = agg['count'].sum(axis=expr.axis, keepdims=expr.keepdims)

    return total / count

@dispatch((std, var))
def _split_chunk(expr, leaf=None, chunk=None, keepdims=True):
    child = expr._subs({leaf: chunk})._child
    return summary(x=child.sum(), x2=(child**2).sum(), n=child.count(),
            keepdims=keepdims, axis=expr.axis)

@dispatch(var)
def _split_agg(expr, leaf=None, agg=None):
    x = agg.x.sum(axis=expr.axis, keepdims=expr.keepdims)
    x2 = agg.x2.sum(axis=expr.axis, keepdims=expr.keepdims)
    n = agg.n.sum(axis=expr.axis, keepdims=expr.keepdims)

    result = (x2 / n) - (x / n)**2

    if expr.unbiased:
        result = result / (n - 1) * n

    return result

@dispatch(std)
def _split_agg(expr, leaf=None, agg=None):
    x = agg.x.sum(axis=expr.axis, keepdims=expr.keepdims)
    x2 = agg.x2.sum(axis=expr.axis, keepdims=expr.keepdims)
    n = agg.n.sum(axis=expr.axis, keepdims=expr.keepdims)

    result = (x2 / n) - (x / n)**2

    if expr.unbiased:
        result = result / (n - 1) * n

    return sqrt(result)

@dispatch(Distinct)
def _split_chunk(expr, leaf=None, chunk=None, **kwargs):
    return expr._subs({leaf: chunk})

@dispatch(Distinct)
def _split_agg(expr, leaf=None, agg=None):
    return agg.distinct()


@dispatch(nunique)
def _split_chunk(expr, leaf=None, chunk=None, **kwargs):
    return (expr._child
                ._subs({leaf: chunk})
                .distinct())

@dispatch(nunique)
def _split_agg(expr, leaf=None, agg=None):
    return agg.distinct().count(keepdims=expr.keepdims)


@dispatch(Summary)
def _split_chunk(expr, leaf=None, chunk=None, keepdims=True):
    exprs = [(name, split(leaf, val, chunk=chunk,
                                       keepdims=False)[0][1])
                            for name, val in zip(expr.fields, expr.values)]
    d = dict()
    for name, e in exprs:
        if isinstance(e, Reduction):
            d[name] = e
        elif isinstance(e, Summary):
            for n, v in zip(e.names, e.values):
                d[name + '_' + n] = v
        else:
            raise NotImplementedError()
    return summary(keepdims=keepdims, **d)


@dispatch(Summary)
def _split_agg(expr, leaf=None, chunk=None, agg=None, keepdims=True):
    exprs = [(name, split(leaf, val, keepdims=False)[1])
                for name, val in zip(expr.fields, expr.values)]

    d = dict()
    for name, (a, ae) in exprs:
        if isscalar(a.dshape.measure): # For simple reductions
            d[name] = ae._subs({a: agg[name]})
        elif isrecord(a.dshape.measure):  # For reductions like mean/var
            names = ['%s_%s' % (name, field) for field in a.fields]
            namedict = dict(zip(a.fields, names))
            d[name] = ae._subs(toolz.merge({a: agg}, namedict))

    return summary(**d)


@dispatch(By)
def _split_chunk(expr, leaf=None, chunk=None, **kwargs):
    chunk_apply = _split_chunk(expr.apply, leaf=leaf, chunk=chunk, keepdims=False)
    chunk_grouper = expr.grouper._subs({leaf: chunk})

    return by(chunk_grouper, chunk_apply)

@dispatch(By)
def _split_agg(expr, leaf=None, agg=None):
    agg_apply = _split_agg(expr.apply, leaf=leaf, agg=agg)
    agg_grouper = expr.grouper._subs({leaf: agg})

    ngroup = len(expr.grouper.fields)

    if isscalar(expr.grouper.dshape.measure):
        agg_grouper = agg[agg.fields[0]]
    else:
        agg_grouper = agg[list(agg.fields[:ngroup])]

    return (by(agg_grouper, agg_apply)
              .relabel(dict(zip(agg.fields[:ngroup],
                                expr.fields[:ngroup]))))


@dispatch((ElemWise, Like, Selection))
def _split_chunk(expr, leaf=None, chunk=None, **kwargs):
    return expr._subs({leaf: chunk})

@dispatch((ElemWise, Like, Selection))
def _split_agg(expr, leaf=None, agg=None):
    return agg


@dispatch(Apply)
def _split_chunk(expr, leaf=None, chunk=None, **kwargs):
    if expr._splittable:
        return expr._subs({leaf: chunk})
    else:
        raise NotImplementedError()

@dispatch(Apply)
def _split_agg(expr, leaf=None, agg=None):
    return agg


from datashape import Fixed
from math import ceil

def dimension_div(a, b):
    """ How many times does b fit into a?

    >>> dimension_div(10, 5)
    2

    We round up
    >>> dimension_div(20, 9)
    3

    In the case of datashape.var, we resort to var
    >>> from datashape import var
    >>> dimension_div(var, 5)
    Var()
    >>> dimension_div(50, var)
    Var()
    """
    if a == datashape.var or b == datashape.var:
        return datashape.var
    if isinstance(a, Fixed):
        a = int(a)
    if isinstance(b, Fixed):
        b = int(b)
    return int(ceil(a / b))


def dimension_mul(a, b):
    """ Given b number of a's how big is our dimension?

    >>> dimension_mul(2, 5)
    10

    We round up
    >>> dimension_mul(9, 3)
    27

    In the case of datashape.var, we resort to var
    >>> from datashape import var
    >>> dimension_mul(datashape.var, 5)
    Var()
    >>> dimension_mul(10, datashape.var)
    Var()
    """
    if a == datashape.var or b == datashape.var:
        return datashape.var
    if isinstance(a, Fixed):
        a = int(a)
    if isinstance(b, Fixed):
        b = int(b)
    return int(a * b)


def aggregate_shape(leaf, expr, chunk, chunk_expr):
    """ The shape of the intermediate aggregate

    >>> leaf = symbol('leaf', '10 * 10 * int')
    >>> expr = leaf.sum(axis=0)
    >>> chunk = symbol('chunk', '3 * 3 * int') # 3 does not divide 10
    >>> chunk_expr = chunk.sum(axis=0, keepdims=True)

    >>> aggregate_shape(leaf, expr, chunk, chunk_expr)
    (4, 10)
    """
    if datashape.var in tconcat(map(shape, [leaf, expr, chunk, chunk_expr])):
        return (datashape.var, ) * leaf.ndim

    numblocks = [int(floor(l / c)) for l, c in zip(leaf.shape, chunk.shape)]
    last_chunk_shape = [l % c for l, c in zip(leaf.shape, chunk.shape)]

    if builtins.sum(last_chunk_shape) != 0:
        old = last_chunk_shape
        last_chunk = symbol(chunk._name,
                            DataShape(*(last_chunk_shape + [chunk.dshape.measure])))
        last_chunk_expr = chunk_expr._subs({chunk: last_chunk})
        last_chunk_shape = shape(last_chunk_expr)
        # Keep zeros if they were there before
        last_chunk_shape = tuple(a if b != 0 else 0 for a, b in
                                zip(last_chunk_shape, old))


    return tuple(int(floor(l / c)) * ce + lce
            for l, c, ce, lce
            in zip(shape(leaf), shape(chunk), shape(chunk_expr), last_chunk_shape))



from __future__ import absolute_import, division, print_function

from collections import Mapping
from keyword import iskeyword
import re

import datashape
from datashape import (
    dshape,
    DataShape,
    Record,
    Var,
    Mono,
    Fixed,
    promote,
    Option,
    Null,
)
from datashape.predicates import isscalar, iscollection, isboolean, isrecord
import numpy as np
from odo.utils import copydoc
import toolz
from toolz import concat, memoize, partial, first
from toolz.curried import map, filter

from ..compatibility import _strtypes, builtins, boundmethod, PY2
from .core import Node, subs, common_subexpression, path, _setattr
from .method_dispatch import select_functions
from ..dispatch import dispatch
from .utils import hashable_index, replace_slices, maxshape
from ..utils import attribute


__all__ = [
    'Apply',
    'Cast',
    'Coalesce',
    'Coerce',
    'ElemWise',
    'Expr',
    'Field',
    'Label',
    'Map',
    'Projection',
    'ReLabel',
    'Selection',
    'SimpleSelection',
    'Slice',
    'Symbol',
    'apply',
    'cast',
    'coalesce',
    'coerce',
    'discover',
    'label',
    'ndim',
    'projection',
    'relabel',
    'selection',
    'shape',
    'symbol',
]


def isvalid_identifier(s):
    """Check whether a string is a valid Python identifier

    Examples
    --------
    >>> isvalid_identifier('Hello')
    True
    >>> isvalid_identifier('Hello world')
    False
    >>> isvalid_identifier('Helloworld!')
    False
    >>> isvalid_identifier('1a')
    False
    >>> isvalid_identifier('a1')
    True
    >>> isvalid_identifier('for')
    False
    >>> isvalid_identifier(None)
    False
    """
    # the re module compiles and caches regexs so no need to compile it
    return (s is not None and not iskeyword(s) and
            re.match(r'^[_a-zA-Z][_a-zA-Z0-9]*$', s) is not None)


def valid_identifier(s):
    """Rewrite a string to be a valid identifier if it contains
    >>> valid_identifier('hello')
    'hello'
    >>> valid_identifier('hello world')
    'hello_world'
    >>> valid_identifier('hello.world')
    'hello_world'
    >>> valid_identifier('hello-world')
    'hello_world'
    >>> valid_identifier(None)
    >>> valid_identifier('1a')
    """
    if isinstance(s, _strtypes):
        if not s or s[0].isdigit():
            return
        return s.replace(' ', '_').replace('.', '_').replace('-', '_')
    return s


class Expr(Node):
    """
    Symbolic expression of a computation

    All Blaze expressions (Join, By, Sort, ...) descend from this class.  It
    contains shared logic and syntax.  It in turn inherits from ``Node`` which
    holds all tree traversal logic
    """

    def _get_field(self, fieldname):
        if not isinstance(self.dshape.measure, (Record, datashape.Map)):
            if fieldname == self._name:
                return self
            raise ValueError(
                "Can not get field '%s' of non-record expression %s" %
                (fieldname, self))
        return Field(self, fieldname)

    def __getitem__(self, key):
        if isinstance(key, _strtypes) and key in self.fields:
            return self._get_field(key)
        elif isinstance(key, Expr) and iscollection(key.dshape):
            return selection(self, key)
        elif (isinstance(key, list) and
              builtins.all(isinstance(k, _strtypes) for k in key)):
            if set(key).issubset(self.fields):
                return self._project(key)
            else:
                raise ValueError('Names %s not consistent with known names %s'
                                 % (key, self.fields))
        elif (isinstance(key, tuple) and
              all(isinstance(k, (int, slice, type(None), list, np.ndarray))
              for k in key)):
            return sliceit(self, key)
        elif isinstance(key, (slice, int, type(None), list, np.ndarray)):
            return sliceit(self, (key,))
        raise ValueError("Not understood %s[%s]" % (self, key))

    def map(self, func, schema=None, name=None):
        return Map(self, func, schema, name)

    def _project(self, key):
        return projection(self, key)

    @attribute
    def schema(self):
        try:
            m = self._schema
        except AttributeError:
            schema = datashape.dshape(self.dshape.measure)
        else:
            schema = m()

        return _setattr(self, 'schema', schema)

    @attribute
    def dshape(self):
        return _setattr(self, 'dshape', self._dshape())

    @property
    def fields(self):
        measure = self.dshape.measure
        if isinstance(self.dshape.measure, Option):
            measure = measure.ty
        if isinstance(measure, Record):
            return measure.names
        elif isinstance(measure, datashape.Map):
            if not isrecord(self.dshape.measure.value):
                raise TypeError('Foreign key must reference a '
                                'Record datashape')
            return measure.value.names
        name = getattr(self, '_name', None)
        if name is not None:
            return [self._name]
        return []

    def _len(self):
        try:
            return int(self.dshape[0])
        except TypeError:
            raise ValueError('Can not determine length of table with the '
                             'following datashape: %s' % self.dshape)

    def __len__(self):  # pragma: no cover
        return self._len()

    def __iter__(self):
        raise NotImplementedError(
            'Iteration over expressions is not supported.\n'
            'Iterate over computed result instead, e.g. \n'
            "\titer(expr)           # don't do this\n"
            "\titer(compute(expr))  # do this instead")

    def __dir__(self):
        result = dir(type(self))
        if (isrecord(self.dshape.measure) or
            isinstance(self.dshape.measure, datashape.Map) and
                self.fields):
            result.extend(map(valid_identifier, self.fields))

        result.extend(toolz.merge(schema_methods(self.dshape.measure),
                                  dshape_methods(self.dshape)))

        return sorted(set(filter(isvalid_identifier, result)))

    def __getattr__(self, key):
        assert key != '_hash', \
            '%s should set _hash in _init' % type(self).__name__
        try:
            result = object.__getattribute__(self, key)
        except AttributeError:
            fields = dict(zip(map(valid_identifier, self.fields), self.fields))

            measure = self.dshape.measure
            if isinstance(measure, datashape.Map): # Foreign key
                measure = measure.key

            # prefer the method if there's a field with the same name
            methods = toolz.merge(
                schema_methods(measure),
                dshape_methods(self.dshape)
            )
            if key in methods:
                func = methods[key]
                if func in method_properties:
                    result = func(self)
                elif getattr(func, '__get__', None):
                    result = func.__get__(self, type(self))
                else:
                    result = boundmethod(func, self)
            elif self.fields and key in fields:
                if isscalar(self.dshape.measure):  # t.foo.foo is t.foo
                    result = self
                else:
                    result = self[fields[key]]
            else:
                raise

        # cache the attribute lookup, getattr will not be invoked again.
        _setattr(self, key, result)
        return result

    @attribute
    def _name(self):
        measure = self.dshape.measure
        if len(self._inputs) == 1 and isscalar(getattr(measure, 'key',
                                                       measure)):
            child_measure = self._child.dshape.measure
            if isscalar(getattr(child_measure, 'key', child_measure)):
                return self._child._name

    def __enter__(self):
        """ Enter context """
        return self

    def __exit__(self, *args):
        """ Exit context

        Close any open resource if we are called in context
        """
        for value in self._resources().values():
            try:
                value.close()
            except AttributeError:
                pass
        return True


def sanitized_dshape(dshape, width=50):
    pretty_dshape = datashape.pprint(dshape, width=width).replace('\n', '')
    if len(pretty_dshape) > width:
        pretty_dshape = "{}...".format(pretty_dshape[:width])
    return pretty_dshape


class Symbol(Expr):
    """
    Symbolic data.  The leaf of a Blaze expression

    Examples
    --------
    >>> points = symbol('points', '5 * 3 * {x: int, y: int}')
    >>> points
    <`points` symbol; dshape='5 * 3 * {x: int32, y: int32}'>
    >>> points.dshape
    dshape("5 * 3 * {x: int32, y: int32}")
    """
    _arguments = '_name', 'dshape', '_token'
    __inputs__ = ()

    def __repr__(self):
        fmt = "<`{}` symbol; dshape='{}'>"
        return fmt.format(self._name, sanitized_dshape(self.dshape))

    def __str__(self):
        return self._name or ''

    def _resources(self):
        return {}


@copydoc(Symbol)
def symbol(name, dshape, token=None):
    return Symbol(name, datashape.dshape(dshape), token or 0)


@dispatch(Symbol, Mapping)
def _subs(o, d):
    """ Subs symbols using symbol function

    Supports caching"""
    newargs = (subs(arg, d) for arg in o._args)
    return symbol(*newargs)


class ElemWise(Expr):
    """
    Elementwise operation.

    The shape of this expression matches the shape of the child.
    """
    def _dshape(self):
        return datashape.DataShape(
            *(self._child.dshape.shape + tuple(self.schema))
        )


class Field(ElemWise):
    """
    A single field from an expression.

    Get a single field from an expression with record-type schema.
    We store the name of the field in the ``_name`` attribute.

    Examples
    --------
    >>> points = symbol('points', '5 * 3 * {x: int32, y: int32}')
    >>> points.x.dshape
    dshape("5 * 3 * int32")

    For fields that aren't valid Python identifiers, use ``[]`` syntax:

    >>> points = symbol('points', '5 * 3 * {"space station": float64}')
    >>> points['space station'].dshape
    dshape("5 * 3 * float64")
    """
    _arguments = '_child', '_name'

    def __str__(self):
        fmt = '%s.%s' if isvalid_identifier(self._name) else '%s[%r]'
        return fmt % (self._child, self._name)

    @property
    def _expr(self):
        return symbol(self._name, datashape.DataShape(self.dshape.measure))

    def _dshape(self):
        shape = self._child.dshape.shape
        measure = self._child.dshape.measure

        # TODO: is this too special-case-y?
        schema = getattr(measure, 'value', measure).dict[self._name]

        shape = shape + schema.shape
        schema = (schema.measure,)
        return DataShape(*(shape + schema))


class Projection(ElemWise):
    """Select a subset of fields from data.

    Examples
    --------
    >>> accounts = symbol('accounts',
    ...                   'var * {name: string, amount: int, id: int}')
    >>> accounts[['name', 'amount']].schema
    dshape("{name: string, amount: int32}")
    >>> accounts[['name', 'amount']]
    accounts[['name', 'amount']]

    See Also
    --------
    blaze.expr.expressions.Field
    """
    _arguments = '_child', '_fields'

    @property
    def fields(self):
        return list(self._fields)

    def _schema(self):
        measure = self._child.schema.measure
        d = getattr(measure, 'value', measure).dict
        return DataShape(Record((name, d[name]) for name in self.fields))

    def __str__(self):
        return '%s[%s]' % (self._child, self.fields)

    def _project(self, key):
        if isinstance(key, list) and set(key).issubset(set(self.fields)):
            return self._child[key]
        raise ValueError("Column Mismatch: %s" % key)

    def _get_field(self, fieldname):
        if fieldname in self.fields:
            return Field(self._child, fieldname)
        raise ValueError("Field %s not found in columns %s" % (fieldname,
                                                               self.fields))


@copydoc(Projection)
def projection(expr, names):
    if not names:
        raise ValueError("Projection with no names")
    if not isinstance(names, (tuple, list)):
        raise TypeError("Wanted list of strings, got %s" % names)
    if not set(names).issubset(expr.fields):
        raise ValueError("Mismatched names. Asking for names %s "
                         "where expression has names %s" %
                         (names, expr.fields))
    return Projection(expr, tuple(names))


def sanitize_index_lists(ind):
    """ Handle lists/arrays of integers/bools as indexes

    >>> sanitize_index_lists([2, 3, 5])
    [2, 3, 5]
    >>> sanitize_index_lists([True, False, True, False])
    [0, 2]
    >>> sanitize_index_lists(np.array([1, 2, 3]))
    [1, 2, 3]
    >>> sanitize_index_lists(np.array([False, True, True]))
    [1, 2]
    """
    if not isinstance(ind, (list, np.ndarray)):
        return ind
    if isinstance(ind, np.ndarray):
        ind = ind.tolist()
    if isinstance(ind, list) and ind and isinstance(ind[0], bool):
        ind = [a for a, b in enumerate(ind) if b]
    return ind


def sliceit(child, index):
    index2 = tuple(map(sanitize_index_lists, index))
    index3 = hashable_index(index2)
    s = Slice(child, index3)
    hash(s)
    return s


class Slice(Expr):
    """Elements `start` until `stop`. On many backends, a `step` parameter
    is also allowed.

    Examples
    --------
    >>> from blaze import symbol
    >>> accounts = symbol('accounts', 'var * {name: string, amount: int}')
    >>> accounts[2:7].dshape
    dshape("5 * {name: string, amount: int32}")
    >>> accounts[2:7:2].dshape
    dshape("3 * {name: string, amount: int32}")
    """
    _arguments = '_child', '_index'

    def _dshape(self):
        return self._child.dshape.subshape[self.index]

    @property
    def index(self):
        return replace_slices(self._index)

    def __str__(self):
        if isinstance(self.index, tuple):
            index = ', '.join(map(str, self._index))
        else:
            index = str(self._index)
        return '%s[%s]' % (self._child, index)


class Selection(Expr):
    """ Filter elements of expression based on predicate

    Examples
    --------

    >>> accounts = symbol('accounts',
    ...                   'var * {name: string, amount: int, id: int}')
    >>> deadbeats = accounts[accounts.amount < 0]
    """
    _arguments = '_child', 'predicate'
    __inputs__ = '_child', 'predicate'

    @property
    def _name(self):
        return self._child._name

    def __str__(self):
        return "%s[%s]" % (self._child, self.predicate)

    def _dshape(self):
        shape = list(self._child.dshape.shape)
        shape[0] = Var()
        return DataShape(*(shape + [self._child.dshape.measure]))


class SimpleSelection(Selection):
    """Internal selection class that does not treat the predicate as an input.
    """
    _arguments = Selection._arguments
    __inputs__ = '_child',


@copydoc(Selection)
def selection(table, predicate):
    subexpr = common_subexpression(table, predicate)

    if not builtins.all(
            isinstance(node, (ElemWise, Symbol)) or
            node.isidentical(subexpr)
            for node in concat([path(predicate, subexpr),
                                path(table, subexpr)])):

        raise ValueError("Selection not properly matched with table:\n"
                         "child: %s\n"
                         "apply: %s\n"
                         "predicate: %s" % (subexpr, table, predicate))

    if not isboolean(predicate.dshape):
        raise TypeError("Must select over a boolean predicate.  Got:\n"
                        "%s[%s]" % (table, predicate))

    return table._subs({subexpr: Selection(subexpr, predicate)})


class Label(ElemWise):
    """An expression with a name.

    Examples
    --------
    >>> accounts = symbol('accounts', 'var * {name: string, amount: int}')
    >>> expr = accounts.amount * 100
    >>> expr._name
    'amount'
    >>> expr.label('new_amount')._name
    'new_amount'

    See Also
    --------
    blaze.expr.expressions.ReLabel
    """
    _arguments = '_child', 'label'

    def _schema(self):
        return self._child.schema

    @property
    def _name(self):
        return self.label

    def _get_field(self, key):
        if key[0] == self.fields[0]:
            return self
        raise ValueError("Column Mismatch: %s" % key)

    def __str__(self):
        return 'label(%s, %r)' % (self._child, self.label)


@copydoc(Label)
def label(expr, lab):
    if expr._name == lab:
        return expr
    return Label(expr, lab)


class ReLabel(ElemWise):
    """
    Table with same content but with new labels

    Examples
    --------
    >>> accounts = symbol('accounts', 'var * {name: string, amount: int}')
    >>> accounts.schema
    dshape("{name: string, amount: int32}")
    >>> accounts.relabel(amount='balance').schema
    dshape("{name: string, balance: int32}")
    >>> accounts.relabel(not_a_column='definitely_not_a_column')
    Traceback (most recent call last):
        ...
    ValueError: Cannot relabel non-existent child fields: {'not_a_column'}
    >>> s = symbol('s', 'var * {"0": int64}')
    >>> s.relabel({'0': 'foo'})
    s.relabel({'0': 'foo'})
    >>> s.relabel(0='foo') # doctest: +SKIP
    Traceback (most recent call last):
        ...
    SyntaxError: keyword can't be an expression

    Notes
    -----
    When names are not valid Python names, such as integers or string with
    spaces, you must pass a dictionary to ``relabel``. For example

    .. code-block:: python

       >>> s = symbol('s', 'var * {"0": int64}')
       >>> s.relabel({'0': 'foo'})
       s.relabel({'0': 'foo'})
       >>> t = symbol('t', 'var * {"whoo hoo": ?float32}')
       >>> t.relabel({"whoo hoo": 'foo'})
       t.relabel({'whoo hoo': 'foo'})

    See Also
    --------
    blaze.expr.expressions.Label
    """
    _arguments = '_child', 'labels'

    def _schema(self):
        subs = dict(self.labels)
        param = self._child.dshape.measure.parameters[0]
        return DataShape(Record([[subs.get(name, name), dtype]
                                 for name, dtype in param]))

    def __str__(self):
        labels = self.labels
        if all(map(isvalid_identifier, map(first, labels))):
            rest = ', '.join('%s=%r' % l for l in labels)
        else:
            rest = '{%s}' % ', '.join('%r: %r' % l for l in labels)
        return '%s.relabel(%s)' % (self._child, rest)


@copydoc(ReLabel)
def relabel(child, labels=None, **kwargs):
    labels = {k: v
              for k, v in toolz.merge(labels or {}, kwargs).items() if k != v}
    label_keys = set(labels)
    fields = child.fields
    if not label_keys.issubset(fields):
        non_existent_fields = label_keys.difference(fields)
        raise ValueError("Cannot relabel non-existent child fields: {%s}" %
                         ', '.join(map(repr, non_existent_fields)))
    if not labels:
        return child
    if isinstance(labels, Mapping):  # Turn dict into tuples
        labels = tuple(sorted(labels.items()))
    if isscalar(child.dshape.measure):
        if child._name == labels[0][0]:
            return child.label(labels[0][1])
        else:
            return child
    return ReLabel(child, labels)


class Map(ElemWise):
    """ Map an arbitrary Python function across elements in a collection

    Examples
    --------
    >>> from datetime import datetime

    >>> t = symbol('t', 'var * {price: real, time: int64}')  # times as integers
    >>> datetimes = t.time.map(datetime.utcfromtimestamp)

    Optionally provide extra schema information

    >>> datetimes = t.time.map(datetime.utcfromtimestamp,
    ...                           schema='{time: datetime}')

    See Also
    --------
    blaze.expr.expresions.Apply
    """
    _arguments = '_child', 'func', '_asschema', '_name0'

    def _schema(self):
        if self._asschema:
            return dshape(self._asschema)
        else:
            raise NotImplementedError("Schema of mapped column not known.\n"
                                      "Please specify datashape keyword in "
                                      ".map method.\nExample: "
                                      "t.columnname.map(function, 'int64')")

    def label(self, name):
        assert isscalar(self.dshape.measure)
        return Map(self._child,
                   self.func,
                   self.schema,
                   name)

    @property
    def shape(self):
        return self._child.shape

    @property
    def ndim(self):
        return self._child.ndim

    @property
    def _name(self):
        if self._name0:
            return self._name0
        else:
            return self._child._name


if PY2:
    copydoc(Map, Expr.map.im_func)
else:
    copydoc(Map, Expr.map)


class Apply(Expr):
    """ Apply an arbitrary Python function onto an expression

    Examples
    --------

    >>> t = symbol('t', 'var * {name: string, amount: int}')
    >>> h = t.apply(hash, dshape='int64')  # Hash value of resultant dataset

    You must provide the datashape of the result with the ``dshape=`` keyword.
    For datashape examples see
    http://datashape.pydata.org/grammar.html#some-simple-examples

    If using a chunking backend and your operation may be safely split and
    concatenated then add the ``splittable=True`` keyword argument

    >>> t.apply(f, dshape='...', splittable=True) # doctest: +SKIP

    See Also
    --------

    blaze.expr.expressions.Map
    """
    _arguments = '_child', 'func', '_asdshape', '_splittable'

    def _schema(self):
        if iscollection(self.dshape):
            return self.dshape.subshape[0]
        else:
            raise TypeError("Non-tabular datashape, %s" % self.dshape)

    def _dshape(self):
        return self._asdshape


@copydoc(Apply)
def apply(expr, func, dshape, splittable=False):
    return Apply(expr, func, datashape.dshape(dshape), splittable)


class Coerce(ElemWise):
    """Coerce an expression to a different type.

    Examples
    --------
    >>> t = symbol('t', '100 * float64')
    >>> t.coerce(to='int64')
    t.coerce(to='int64')
    >>> t.coerce('float32')
    t.coerce(to='float32')
    >>> t.coerce('int8').dshape
    dshape("100 * int8")
    """
    _arguments = '_child', 'to'

    def _schema(self):
        return self.to

    def __str__(self):
        return '%s.coerce(to=%r)' % (self._child, str(self.schema))


@copydoc(Coerce)
def coerce(expr, to):
    return Coerce(expr, dshape(to) if isinstance(to, _strtypes) else to)


class Cast(Expr):
    """Cast an expression to a different type.

    This is only an expression time operation.

    Examples
    --------
    >>> s = symbol('s', '?int64')
    >>> s.cast('?int32').dshape
    dshape("?int32")

    # Cast to correct mislabeled optionals
    >>> s.cast('int64').dshape
    dshape("int64")

    # Cast to give concrete dimension length
    >>> t = symbol('t', 'var * float32')
    >>> t.cast('10 * float32').dshape
    dshape("10 * float32")
    """
    _arguments = '_child', 'to'

    def _dshape(self):
        return self.to

    def __str__(self):
        return 'cast(%s, to=%r)' % (self._child, str(self.to))


@copydoc(Cast)
def cast(expr, to):
    return Cast(expr, dshape(to) if isinstance(to, _strtypes) else to)


Expr.cast = cast  # method of all exprs


def binop_name(expr):
    if not isscalar(expr.dshape.measure):
        return None
    l = getattr(expr.lhs, '_name', None)
    r = getattr(expr.rhs, '_name', None)
    if bool(l) ^ bool(r):
        return l or r
    elif l == r:
        return l

    return None


def binop_inputs(expr):
    if isinstance(expr.lhs, Expr):
        yield expr.lhs
    if isinstance(expr.rhs, Expr):
        yield expr.rhs


class Coalesce(Expr):
    """SQL like coalesce.

    coalesce(a, b) = {
        a if a is not NULL
        b otherwise
    }

    Examples
    --------
    >>> coalesce(1, 2)
    1
    >>> coalesce(1, None)
    1
    >>> coalesce(None, 2)
    2
    >>> coalesce(None, None) is None
    True
    """
    _arguments = 'lhs', 'rhs', 'dshape'
    __inputs__ = 'lhs', 'rhs'

    def __str__(self):
        return 'coalesce(%s, %s)' % (self.lhs, self.rhs)

    _name = property(binop_name)

    @property
    def _inputs(self):
        return tuple(binop_inputs(self))


@copydoc(Coalesce)
def coalesce(a, b):
    a_dshape = discover(a)
    a_measure = a_dshape.measure
    isoption = isinstance(a_measure, Option)
    if isoption:
        a_measure = a_measure.ty
    isnull = isinstance(a_measure, Null)
    if isnull:
        # a is always null, this is just b
        return b

    if not isoption:
        # a is not an option, this is just a
        return a

    b_dshape = discover(b)
    return Coalesce(a, b, DataShape(*(
        maxshape((a_dshape.shape, b_dshape.shape)) +
        (promote(a_measure, b_dshape.measure),)
    )))


dshape_method_list = list()
schema_method_list = list()
method_properties = set()

dshape_methods = memoize(partial(select_functions, dshape_method_list))
schema_methods = memoize(partial(select_functions, schema_method_list))


@dispatch(DataShape)
def shape(ds):
    s = ds.shape
    s = tuple(int(d) if isinstance(d, Fixed) else d for d in s)
    return s


@dispatch(object)
def shape(expr):
    """ Shape of expression

    >>> symbol('s', '3 * 5 * int32').shape
    (3, 5)

    Works on anything discoverable

    >>> shape([[1, 2], [3, 4]])
    (2, 2)
    """
    s = list(discover(expr).shape)
    for i, elem in enumerate(s):
        try:
            s[i] = int(elem)
        except TypeError:
            pass

    return tuple(s)


def ndim(expr):
    """ Number of dimensions of expression

    >>> symbol('s', '3 * var * int32').ndim
    2
    """
    return len(shape(expr))


dshape_method_list.extend([
    (lambda ds: True, set([apply])),
    (iscollection, set([shape, ndim])),
    (lambda ds: iscollection(ds) and isscalar(ds.measure), set([coerce]))
])

schema_method_list.extend([
    (isscalar, set([label, relabel, coerce])),
    (isrecord, set([relabel])),
    (lambda ds: isinstance(ds, Option), {coalesce}),
])

method_properties.update([shape, ndim])


@dispatch(Expr)
def discover(expr):
    return expr.dshape



import pytest

import os

from blaze import data, compute
from blaze.utils import raises
from odo import URL, CSV

import pandas as pd
import pandas.util.testing as tm

from functools import partial

try:
    from urllib2 import urlopen
    from urllib2 import HTTPError, URLError
except ImportError:
    from urllib.request import urlopen
    from urllib.error import HTTPError, URLError

pytestmark = pytest.mark.skipif(raises(URLError,
                                       partial(urlopen, "http://google.com")),
                                reason='unable to connect to google.com')

iris_url = ('https://raw.githubusercontent.com/'
            'blaze/blaze/master/blaze/examples/data/iris.csv')

@pytest.fixture
def iris_local():
    thisdir = os.path.abspath(os.path.dirname(__file__))
    return data(os.path.join(thisdir, os.pardir, os.pardir, "examples", "data", "iris.csv"))

def test_url_csv_data(iris_local):
    iris_remote = data(iris_url)
    assert isinstance(iris_remote.data, URL(CSV))
    iris_remote_df = compute(iris_remote)
    assert isinstance(iris_remote_df, pd.DataFrame)
    iris_local_df = compute(iris_local)
    tm.assert_frame_equal(iris_remote_df, iris_local_df)



from __future__ import absolute_import, division, print_function

import datetime
from functools import partial

from datashape import Mono, DataShape
import pandas as pd

from blaze.dispatch import dispatch
from blaze.compatibility import unicode


json_dumps_ns = dict()
dispatch = partial(dispatch, namespace=json_dumps_ns)


@dispatch(datetime.datetime)
def json_dumps(dt):
    if dt is pd.NaT:
        # NaT has an isoformat but it is totally invalid.
        # This keeps the parsing on the client side simple.
        s = u'NaT'
    else:
        s = dt.isoformat()
        if isinstance(s, bytes):
            s = s.decode('utf-8')
        if not dt.tzname():
            s += u'Z'

    return {u'__!datetime': s}


@dispatch(frozenset)
def json_dumps(ds):
    return {u'__!frozenset': list(ds)}


@dispatch(datetime.timedelta)
def json_dumps(ds):
    return {u'__!timedelta': ds.total_seconds()}


@dispatch(Mono)
def json_dumps(m):
    return {u'__!mono': unicode(m)}


@dispatch(DataShape)
def json_dumps(ds):
    return {u'__!datashape': unicode(ds)}


@dispatch(object)
def json_dumps(ob):
    return pd.io.packers.encode(ob)



from __future__ import absolute_import, division, print_function

import pandas
import os
from toolz import curry, concat
import pandas as pd
import numpy as np
from collections import Iterator, Iterable
from odo import into, Temp
from odo.backends.csv import CSV
from odo.backends.url import URL
from multipledispatch import MDNotImplementedError
import dask.dataframe as dd

from ..dispatch import dispatch
from ..expr import Expr, Head, ElemWise, Distinct, Symbol, Projection, Field
from ..expr.core import path
from ..utils import available_memory
from ..expr.split import split
from .core import compute
from ..expr.optimize import lean_projection
from .pmap import get_default_pmap
from warnings import warn


__all__ = ['optimize', 'pre_compute']


@dispatch(Expr, CSV)
def optimize(expr, _):
    return lean_projection(expr)  # This is handled in pre_compute


@dispatch(Expr, CSV)
def pre_compute(expr, data, comfortable_memory=None, chunksize=None, blocksize=None, **kwargs):
    comfortable_memory = comfortable_memory or min(1e9, available_memory() / 4)

    kwargs = dict()

    # Chunk if the file is large
    if os.path.getsize(data.path) > comfortable_memory:
        do_chunk = True
        if chunksize is not None:
            warn("Deprecation warning: chunksize keyword renamed to blocksize")
            blocksize = chunksize
        if blocksize is not None:
            kwargs['blocksize'] = blocksize
    else:
        do_chunk = False

    # Insert projection into read_csv
    oexpr = optimize(expr, data)
    leaf = oexpr._leaves()[0]
    pth = list(path(oexpr, leaf))
    if len(pth) >= 2 and isinstance(pth[-2], (Projection, Field)):
        # NOTE: FIXME: We pass the column names through `str` to workaround a
        # PY2 Pandas bug with strings / unicode objects.
        kwargs['usecols'] = list(map(str, pth[-2].fields))

    if do_chunk:
        return dd.read_csv(data.path, **kwargs)
    else:
        return into(pd.DataFrame, data, dshape=leaf.dshape, **kwargs)


@dispatch((Expr, Head), URL(CSV))
def pre_compute(expr, data, **kwargs):
    return pre_compute(expr, into(Temp(CSV), data, **kwargs), **kwargs)


Cheap = (Head, ElemWise, Distinct, Symbol)

@dispatch(Head, CSV)
def pre_compute(expr, data, **kwargs):
    leaf = expr._leaves()[0]
    if all(isinstance(e, Cheap) for e in path(expr, leaf)):
        return into(Iterator, data, chunksize=10000, dshape=leaf.dshape)
    else:
        raise MDNotImplementedError()



from __future__ import absolute_import, division, print_function

import pytest
pyspark = pytest.importorskip('pyspark')

import pandas as pd
from blaze import compute, symbol, summary, exp, by, join, merge
from toolz import identity


data = [['Alice', 100, 1],
        ['Bob', 200, 2],
        ['Alice', 50, 3]]

data2 = [['Alice', 'Austin'],
         ['Bob', 'Boston']]

df = pd.DataFrame(data, columns=['name', 'amount', 'id'])


# this only exists because we need to have a single session scoped spark
# context, otherwise these would simply be global variables
@pytest.fixture
def rdd(sc):
    return sc.parallelize(data)


@pytest.fixture
def rdd2(sc):
    return sc.parallelize(data2)


t = symbol('t', 'var * {name: string, amount: int, id: int}')

t2 = symbol('t2', 'var * {name: string, city: string}')

# Web Commons Graph Example data
data_idx = [['A', 1],
            ['B', 2],
            ['C', 3]]

data_arc = [[1, 3],
            [2, 3],
            [3, 1]]

t_idx = symbol('idx', 'var * {name: string, node_id: int32}')

t_arc = symbol('arc', 'var * {node_out: int32, node_id: int32}')


def test_symbol(rdd):
    assert compute(t, rdd) == rdd


def test_projection(rdd):
    assert compute(t['name'], rdd).collect() == [row[0] for row in data]


def test_multicols_projection(rdd):
    result = compute(t[['amount', 'name']], rdd).collect()
    expected = [(100, 'Alice'), (200, 'Bob'), (50, 'Alice')]

    print(result)
    print(expected)

    assert result == expected


reduction_exprs = (t['amount'].sum(),
                   t['amount'].min(),
                   t['amount'].max(),
                   t['amount'].nunique(),
                   t['name'].nunique(),
                   t['amount'].count(),
                   (t['amount'] > 150).any(),
                   (t['amount'] > 150).all(),
                   t['amount'].mean(),
                   t['amount'].var(),
                   summary(a=t.amount.sum(), b=t.id.count()),
                   t['amount'].std())


@pytest.mark.parametrize('expr', reduction_exprs)
def test_reductions(expr, rdd):
    result = compute(expr, rdd)
    expected = compute(expr, data)
    if not result == expected:
        print(result)
        print(expected)
        if isinstance(result, float):
            assert abs(result - expected) < 0.001
        else:
            assert result == expected


exprs = (t['amount'],
         t['amount'] == 100,
         t['amount'].truncate(150),
         t[t['name'] == 'Alice'],
         t[t['amount'] == 0],
         t[t['amount'] > 150],
         t['amount'] + t['id'],
         t['amount'] % t['id'],
         exp(t['amount']),
         by(t['name'], total=t['amount'].sum()),
         by(t['name'], total=(t['amount'] + 1).sum()),
         (t['amount'] * 1).label('foo'),
         t.map(lambda tup: tup[1] + tup[2], 'real'),
         t[t.name.like('Alice')],
         t['amount'].apply(identity, 'var * real', splittable=True),
         t['amount'].map(lambda x: x + 1, 'int'))

exprs = list(zip(map(str, exprs), exprs))


def tuplify(x):
    return tuple(x) if isinstance(x, list) else x


@pytest.mark.parametrize(['string', 'expr'], exprs)
def test_basic(rdd, string, expr):
    result = set(map(tuplify, compute(expr, rdd).collect()))
    expected = set(map(tuplify, compute(expr, data)))
    assert result == expected


tbig = symbol(
    'tbig', 'var * {name: string, sex: string[1], amount: int, id: int}')

big_exprs = [
    by(tbig[['name', 'sex']], total=tbig['amount'].sum()),
    by(tbig[['name', 'sex']], total=(tbig['id'] + tbig['amount']).sum())]


@pytest.mark.parametrize('expr', big_exprs)
def test_big_by(sc, expr):
    data = [['Alice', 'F', 100, 1],
            ['Alice', 'F', 100, 3],
            ['Drew', 'F', 100, 4],
            ['Drew', 'M', 100, 5],
            ['Drew', 'M', 200, 5]]
    rdd = sc.parallelize(data)
    result = set(map(tuplify, compute(expr, rdd).collect()))
    expected = set(map(tuplify, compute(expr, data)))
    assert result == expected


def test_head(rdd):
    assert list(compute(t.head(1), rdd)) == list(compute(t.head(1), data))


sort_exprs = [
    t.sort('amount'),
    t.sort('amount', ascending=True),
    t.sort(t['amount'], ascending=True),
    t.sort(-t['amount'].label('foo') + 1, ascending=True),
    t.sort(['amount', 'id'])
]


@pytest.mark.parametrize('expr', sort_exprs)
def test_sort(rdd, expr):
    result = compute(expr, rdd).collect()
    expected = list(compute(expr, data))
    assert result == expected


def test_distinct(rdd):
    assert set(compute(t['name'].distinct(), rdd).collect()) == \
        set(['Alice', 'Bob'])


@pytest.mark.xfail(
    raises=NotImplementedError,
    reason='cannot specify columns to distinct on yet',
)
def test_distinct_on(rdd):
    compute(t.distinct('name'), rdd)


def test_join(rdd, rdd2):

    joined = join(t, t2, 'name')
    expected = [('Alice', 100, 1, 'Austin'),
                ('Bob', 200, 2, 'Boston'),
                ('Alice', 50, 3, 'Austin')]
    result = compute(joined, {t: rdd, t2: rdd2}).collect()
    assert all(i in expected for i in result)


def test_multi_column_join(sc):
    left = [(1, 2, 3),
            (2, 3, 4),
            (1, 3, 5)]
    right = [(1, 2, 30),
             (1, 3, 50),
             (1, 3, 150)]
    rleft = sc.parallelize(left)
    rright = sc.parallelize(right)

    L = symbol('L', 'var * {x: int, y: int, z: int}')
    R = symbol('R', 'var * {x: int, y: int, w: int}')

    j = join(L, R, ['x', 'y'])

    result = compute(j, {L: rleft, R: rright})
    expected = [(1, 2, 3, 30),
                (1, 3, 5, 50),
                (1, 3, 5, 150)]

    assert set(result.collect()) == set(expected)


def test_groupby(sc):
    rddidx = sc.parallelize(data_idx)
    rddarc = sc.parallelize(data_arc)

    joined = join(t_arc, t_idx, "node_id")

    t = by(joined['name'], count=joined['node_id'].count())
    a = compute(t, {t_arc: rddarc, t_idx: rddidx})
    in_degree = dict(a.collect())
    assert in_degree == {'A': 1, 'C': 2}


def test_multi_level_rowfunc_works(rdd):
    expr = t['amount'].map(lambda x: x + 1, 'int')

    assert compute(expr, rdd).collect() == [x[1] + 1 for x in data]


def test_merge(rdd):
    col = (t['amount'] * 2).label('new')
    expr = merge(t['name'], col)

    assert compute(expr, rdd).collect() == [
        (row[0], row[1] * 2) for row in data]


def test_selection_out_of_order(rdd):
    expr = t['name'][t['amount'] < 100]

    assert compute(expr, rdd).collect() == ['Alice']


def test_recursive_rowfunc_is_used(rdd):
    expr = by(t['name'], total=(2 * (t['amount'] + t['id'])).sum())
    expected = [('Alice', 2 * (101 + 53)),
                ('Bob', 2 * (202))]
    assert set(compute(expr, rdd).collect()) == set(expected)


def test_outer_join(sc):
    left = [(1, 'Alice', 100),
            (2, 'Bob', 200),
            (4, 'Dennis', 400)]
    left = sc.parallelize(left)
    right = [('NYC', 1),
             ('Boston', 1),
             ('LA', 3),
             ('Moscow', 4)]
    right = sc.parallelize(right)

    L = symbol('L', 'var * {id: int, name: string, amount: real}')
    R = symbol('R', 'var * {city: string, id: int}')

    assert set(compute(join(L, R), {L: left, R: right}).collect()) == set(
        [(1, 'Alice', 100, 'NYC'),
         (1, 'Alice', 100, 'Boston'),
         (4, 'Dennis', 400, 'Moscow')])

    assert set(compute(join(L, R, how='left'), {L: left, R: right}).collect()) == set(
        [(1, 'Alice', 100, 'NYC'),
         (1, 'Alice', 100, 'Boston'),
         (2, 'Bob', 200, None),
         (4, 'Dennis', 400, 'Moscow')])

    assert set(compute(join(L, R, how='right'), {L: left, R: right}).collect()) == set(
        [(1, 'Alice', 100, 'NYC'),
         (1, 'Alice', 100, 'Boston'),
         (3, None, None, 'LA'),
         (4, 'Dennis', 400, 'Moscow')])

    # Full outer join not yet supported
    assert set(compute(join(L, R, how='outer'), {L: left, R: right}).collect()) == set(
        [(1, 'Alice', 100, 'NYC'),
         (1, 'Alice', 100, 'Boston'),
         (2, 'Bob', 200, None),
         (3, None, None, 'LA'),
         (4, 'Dennis', 400, 'Moscow')])



from __future__ import absolute_import, division, print_function

import datetime

try:
    from cytoolz import curry
except ImportError:
    from toolz import curry

from datashape import dshape
import numpy as np
import pandas as pd
from pandas.io.packers import decode

# Imports that replace older utils.
from blaze.compatibility import PY2

# dict of converters. This is stored as a default arg to object hook for
# performance because this function is really really slow when unpacking data.
# This is a mutable default but it is not a bug!
_converters = {}


if PY2:
    _keys = dict.keys
else:
    def _keys(d, _dkeys=dict.keys, _list=list):
        return _list(_dkeys(d))


def object_hook(ob,
                # Cached for performance. Forget these exist.
                _len=len,
                _keys=_keys,
                _first_three_chars=np.s_[:3],
                _converters=_converters):
    """Convert a json object dict back into a python object.

    This looks for our objects that have encoded richer representations
    with a ``__!{type}`` key.

    Parameters
    ----------
    ob : dict
        The raw json parsed dictionary.

    Returns
    -------
    parsed : any
        The richer form of the object.

    Notes
    -----
    The types that this reads can be extended with the ``register`` method.
    For example:

    >>> class MyList(list):
    ...     pass
    >>> @object_hook.register('MyList')
    ... def _parse_my_list(ob):
    ...     return MyList(ob)

    Register can also be called as a function like:
    >>> a = object_hook.register('frozenset', frozenset)
    >>> a is frozenset
    True
    """
    if _len(ob) != 1:
        return decode(ob)

    key = _keys(ob)[0]
    if key[_first_three_chars] != '__!':
        return ob

    return _converters[key](ob[key])


@curry
def register(typename, converter, converters=_converters):
    converters['__!' + typename] = converter
    return converter
object_hook.register = register

object_hook._converters = _converters  # make this accesible for debugging
del _converters
del _keys  # captured by default args


object_hook.register('datetime', pd.Timestamp)
object_hook.register('frozenset', frozenset)
object_hook.register('datashape', dshape)


@object_hook.register('mono')
def _read_mono(m):
    return dshape(m).measure


@object_hook.register('timedelta')
def _read_timedelta(ds):
    return datetime.timedelta(seconds=ds)


@object_hook.register('bytes')
def _read_bytes(bs):
    if not isinstance(bs, bytes):
        bs = bs.encode('latin1')
    return bs



from __future__ import absolute_import, division, print_function
from copy import deepcopy

from datashape.predicates import isscalar
from multipledispatch import MDNotImplementedError

from .expressions import *
from .strings import *
from .arithmetic import *
from .collections import *
from .split_apply_combine import *
from .broadcast import *
from .reductions import *
from ..dispatch import dispatch


def lean_projection(expr):
    """ Insert projections to keep dataset as thin as possible

    >>> t = symbol('t', 'var * {a: int, b: int, c: int, d: int}')
    >>> lean_projection(t.sort('a').b)
    t[['a', 'b']].sort('a', ascending=True).b
    """
    fields = expr.fields
    return _lean(expr, fields=fields)[0]


@dispatch(Symbol)
def _lean(expr, fields=None):
    """

    >>> s = symbol('s', '{x: int, y: int}')
    >>> _lean(s, ('x',))
    (s['x'], ('x',))

    >>> _lean(s, ())
    (s, ())

    >>> s = symbol('s', 'int')
    >>> _lean(s, ())
    (s, ())
    >>> _lean(s, ('s',))
    (s, ())
    """
    if not fields or set(expr.fields).issubset(fields):
        return expr, fields
    else:
        return expr[sorted(fields)], fields


@dispatch(Projection)
def _lean(expr, fields=None):
    child, _ = _lean(expr._child, fields=fields)
    return child[sorted(fields, key=expr.fields.index)], fields


@dispatch(Field)
def _lean(expr, fields=None):
    fields = set(fields)
    fields.add(expr._name)

    child, _ = _lean(expr._child, fields=fields)
    return child[expr._name], fields


@dispatch(Arithmetic)
def _lean(expr, fields=None):
    lhs, right_fields = _lean(expr.lhs, fields=())
    rhs, left_fields = _lean(expr.rhs, fields=())
    new_fields = set(fields) | set(left_fields) | set(right_fields)

    return type(expr)(lhs, rhs), new_fields


@dispatch(object)
def _lean(expr, fields=None):
    return expr, fields


@dispatch(Label)
def _lean(expr, fields=None):
    child, new_fields = _lean(expr._child, fields=())
    return child.label(expr._name), new_fields


@dispatch(ReLabel)
def _lean(expr, fields=None):
    labels = dict(expr.labels)
    reverse_labels = dict((v, k) for k, v in expr.labels)

    child_fields = set(reverse_labels.get(f, f) for f in fields)

    child, new_fields = _lean(expr._child, fields=child_fields)
    return child.relabel(**dict((k, v) for k, v in expr.labels if k in
        child.fields)), new_fields


@dispatch(ElemWise)
def _lean(expr, fields=None):
    if isscalar(expr._child.dshape.measure):
        child, _ = _lean(expr._child, fields=set(expr._child.fields))
        return expr._subs({expr._child: child}), set(expr._child.fields)
    else:
        raise MDNotImplementedError()


@dispatch(Broadcast)
def _lean(expr, fields=None):
    fields = set(fields) | set(expr.active_columns())

    child, _ = _lean(expr._child, fields=fields)
    return expr._subs({expr._child: child}), fields


@dispatch(Selection)
def _lean(expr, fields=None):
    predicate, pred_fields = _lean(expr.predicate, fields=fields)

    fields = set(fields) | set(pred_fields)

    child, _ = _lean(expr._child, fields=fields)
    return expr._subs({expr._child: child}), fields


@dispatch(Sort)
def _lean(expr, fields=None):
    key = expr.key
    if not isinstance(key, (list, set, tuple)):
        key = [key]
    new_fields = set(fields) | set(key)

    child, _ = _lean(expr._child, fields=new_fields)
    return child.sort(key=expr.key, ascending=expr.ascending), new_fields


@dispatch(Head)
def _lean(expr, fields=None):
    child, child_fields = _lean(expr._child, fields=fields)
    return child.head(expr.n), child_fields


@dispatch(Reduction)
def _lean(expr, fields=None):
    child = expr._child
    try:
        fields = child.active_columns()
    except AttributeError:
        fields = child.fields
    child, child_fields = _lean(child, fields=set(filter(None, fields)))
    return expr._subs({expr._child: child}), child_fields


@dispatch(Summary)
def _lean(expr, fields=None):
    save = dict()
    new_fields = set()
    for name, val in zip(expr.names, expr.values):
        if name not in fields:
            continue
        child, child_fields = _lean(val, fields=set())
        save[name] = child
        new_fields |= set(child_fields)

    return summary(**save), new_fields


@dispatch(By)
def _lean(expr, fields=None):
    fields = set(fields)
    grouper, grouper_fields = _lean(expr.grouper,
                                    fields=fields.intersection(expr.grouper.fields))
    apply, apply_fields = _lean(expr.apply,
                                fields=fields.intersection(expr.apply.fields))

    new_fields = set(apply_fields) | set(grouper_fields)

    child = common_subexpression(grouper, apply)
    if len(child.fields) > len(new_fields):
        child, _ = _lean(child, fields=new_fields)
        grouper = grouper._subs({expr._child: child})
        apply = apply._subs({expr._child: child})

    return By(grouper, apply), new_fields


@dispatch(Distinct)
def _lean(expr, fields=None):
    child, new_fields = _lean(expr._child, fields=expr.fields)
    return expr._subs({expr._child: child}), new_fields


@dispatch(Merge)
def _lean(expr, fields=None):
    new_fields = set()
    for f in expr.fields:
        if f not in fields:
            continue
        le, nf = _lean(expr[f], fields=set([f]))
        new_fields.update(nf)
    child, _ = _lean(expr._child, fields=new_fields)

    return expr._subs({expr._child: child})[sorted(fields)], new_fields


@dispatch((Join, Concat))
def _lean(expr, fields=None):
    return expr, fields


@dispatch(Expr)
def _lean(expr, fields=None):
    """ Lean projection version of expression

    Parameters
    ----------

    expr : Expression
        An expression to be optimized
    fields : Iterable of strings
        The fields that will be needed from this expression

    Returns
    -------

    expr : Expression
        An expression with Projections inserted to avoid unnecessary fields
    fields : Iterable of strings
        The fields that this expression requires to execute
    """
    raise NotImplementedError()


@dispatch(Selection)
def simple_selections(expr):
    """Cast all ``Selection`` nodes into ``SimpleSelection`` nodes.

    This causes the compute core to not treat the predicate as an input.

    Parameters
    ----------
    expr : Expr
        The expression to traverse.

    Returns
    -------
    siplified : Expr
        The expression with ``Selection``s replaces with ``SimpleSelection``s.
    """
    return SimpleSelection(
        simple_selections(expr._child),
        simple_selections(expr.predicate),
    )


@dispatch(Expr)
def simple_selections(expr):
    return expr._subs({e: simple_selections(e) for e in expr._inputs})



from blaze.expr.utils import *
from blaze.expr.utils import _slice


def test__slice_object():
    s = _slice(1, 10, 2)
    assert str(s) == '1:10:2'
    assert hash(s)

    assert _slice(1, 10, 2) == _slice(1, 10, 2)



from blaze.expr import Expr, symbol, Cast
from blaze.compute import compute_up
from blaze import compute


class Foo(object):
    def __init__(self, data):
        self.data = data


@compute_up.register((Expr, Cast), Foo)
def compute_up(expr, data, **kwargs):
    return data


def optimize(expr, data):
    """ Renames leaf """
    leaf = expr._leaves()[0]
    return expr._subs({leaf: symbol('newname', leaf.dshape)})


def test_scope_gets_updated_after_optimize_call():
    a = symbol('a', 'int')
    result = compute(a + 1, Foo('foo'), optimize=optimize)
    assert result.data == 'foo'



from __future__ import absolute_import, division, print_function

from .core import *
from .expressions import *
from .strings import *
from .datetime import *
from .math import *
from .arithmetic import *
from .broadcast import *
from .collections import *
from .split_apply_combine import *
from .reductions import *
from .arrays import *
from . import math



from __future__ import absolute_import, division, print_function

from collections import namedtuple

from .core import pre_compute
from ..expr import Expr, Field
from ..dispatch import dispatch

from odo import into, chunks

import pandas as pd


@dispatch(Expr, pd.io.pytables.AppendableFrameTable)
def pre_compute(expr, data, **kwargs):
    return into(chunks(pd.DataFrame), data, **kwargs)


@dispatch(Expr, pd.io.pytables.Fixed)
def pre_compute(expr, data, **kwargs):
    return into(pd.DataFrame, data, **kwargs)


@dispatch(Field, pd.HDFStore)
def compute_up(expr, data, **kwargs):
    key = '/' + expr._name
    if key in data.keys():
        return data.get_storer(key)
    else:
        return HDFGroup(data, key)


HDFGroup = namedtuple('HDFGroup', 'parent,datapath')


@dispatch(Field, HDFGroup)
def compute_up(expr, data, **kwargs):
    key = data.datapath + '/' + expr._name
    if key in data.parent.keys():
        return data.parent.get_storer(key)
    else:
        return HDFGroup(data.parent, key)



from blaze.expr.method_dispatch import select_functions, name, partial


def iseven(n):
    return isinstance(n, int) and n % 2 == 0

def inc(x):
    return Foo(x.data + 1)

def dec(x):
    return Foo(x.data - 1)

def lower(x):
    return Foo(x.data.lower())

def upper(x):
    return Foo(x.data.upper())

def halve(x):
    return Foo(x.data // 2)

def isnine(x):
    return True


methods = [(int, set([inc, dec])),
           (str, set([lower, upper])),
           (iseven, set([halve])),
           (9, set([isnine]))]


def test_select_functions():
    assert select_functions(methods, 3) == {'inc': inc, 'dec': dec}
    assert select_functions(methods, 4) == {'inc': inc, 'dec': dec, 'halve': halve}
    assert select_functions(methods, 'A') == {'lower': lower, 'upper': upper}
    assert select_functions(methods, 9) == {'inc': inc, 'dec': dec, 'isnine':
        isnine}


def test_name():
    assert name(inc) == 'inc'
    assert name(partial(inc)) == name(inc)
